/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPattern.cpp
    \brief    neighbouring pixel access classes
*/

#include "TComPic.h"
#include "TComPattern.h"
#include "TComDataCU.h"
#include "TComTU.h"
#include "Debug.h"
#include "TComPrediction.h"
#include "cmath"
//! \ingroup TLibCommon
//! \{

// Forward declarations

/// padding of unavailable reference samples for intra prediction
Void fillReferenceSamples( const Int bitDepth, TComDataCU* pcCU, const Pel* piRoiOrigin, Pel* piAdiTemp, const Bool* bNeighborFlags,
                           const Int iNumIntraNeighbor, const Int unitWidth, const Int unitHeight, const Int iAboveUnits, const Int iLeftUnits,
                           const UInt uiCuWidth, const UInt uiCuHeight, const UInt uiWidth, const UInt uiHeight, const Int iPicStride,
                           const ChannelType chType, const ChromaFormat chFmt );

/// constrained intra prediction
Bool  isAboveLeftAvailable  ( TComDataCU* pcCU, UInt uiPartIdxLT );
Int   isAboveAvailable      ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool* bValidFlags );
Int   isLeftAvailable       ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool* bValidFlags );
Int   isAboveRightAvailable ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool* bValidFlags );
Int   isBelowLeftAvailable  ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool* bValidFlags );


// ====================================================================================================================
// Public member functions (TComPatternParam)
// ====================================================================================================================

/** \param  piTexture     pixel data
 \param  iRoiWidth     pattern width
 \param  iRoiHeight    pattern height
 \param  iStride       buffer stride
 \param  iOffsetLeft   neighbour offset (left)
 \param  iOffsetRight  neighbour offset (right)
 \param  iOffsetAbove  neighbour offset (above)
 \param  iOffsetBottom neighbour offset (bottom)
 */
Void TComPatternParam::setPatternParamPel ( Pel* piTexture,
                                           Int iRoiWidth,
                                           Int iRoiHeight,
                                           Int iStride
                                           )
{
  m_piROIOrigin    = piTexture;
  m_iROIWidth       = iRoiWidth;
  m_iROIHeight      = iRoiHeight;
  m_iPatternStride  = iStride;
}

// ====================================================================================================================
// Public member functions (TComPattern)
// ====================================================================================================================

Void TComPattern::initPattern (Pel* piY,
                               Int iRoiWidth,
                               Int iRoiHeight,
                               Int iStride)
{
  m_cPatternY. setPatternParamPel( piY,  iRoiWidth, iRoiHeight, iStride);
}


// NOTE: RExt - this has been kept in this C++ file to allow easier tracking/comparison against HM.
Void TComPrediction::initAdiPatternChType( TComTU &rTu, Bool& bAbove, Bool& bLeft, const ComponentID compID, const Bool bFilterRefSamples DEBUG_STRING_FN_DECLARE(sDebug))
{
  // NOTE: RExt - This function has been modified for increased flexibility as part of the Square TU 4:2:2 implementation
  const ChannelType chType    = toChannelType(compID);

  TComDataCU *pcCU=rTu.getCU();
  const UInt uiZorderIdxInPart=rTu.GetAbsPartIdxTU();
  const UInt uiTuWidth        = rTu.getRect(compID).width;
  const UInt uiTuHeight       = rTu.getRect(compID).height;
  const UInt uiTuWidth2       = uiTuWidth  << 1;
  const UInt uiTuHeight2      = uiTuHeight << 1;

  const Int  iBaseUnitSize    = g_uiMaxCUWidth >> g_uiMaxCUDepth;
  const Int  iUnitWidth       = iBaseUnitSize  >> pcCU->getPic()->getPicYuvRec()->getComponentScaleX(compID);
  const Int  iUnitHeight      = iBaseUnitSize  >> pcCU->getPic()->getPicYuvRec()->getComponentScaleY(compID);
  const Int  iTUWidthInUnits  = uiTuWidth  / iUnitWidth;
  const Int  iTUHeightInUnits = uiTuHeight / iUnitHeight;
  const Int  iAboveUnits      = iTUWidthInUnits  << 1;
  const Int  iLeftUnits       = iTUHeightInUnits << 1;

  const Int  iPartIdxStride   = pcCU->getPic()->getNumPartInWidth(); //NOTE: RExt - despite the name "TComPic::getNumPartInWidth", this gets the number of parts in an **LCU** width, not the picture width
  const UInt uiPartIdxLT      = pcCU->getZorderIdxInCU() + uiZorderIdxInPart;
  const UInt uiPartIdxRT      = g_auiRasterToZscan[ g_auiZscanToRaster[ uiPartIdxLT ] +   iTUWidthInUnits  - 1                   ];
  const UInt uiPartIdxLB      = g_auiRasterToZscan[ g_auiZscanToRaster[ uiPartIdxLT ] + ((iTUHeightInUnits - 1) * iPartIdxStride)];

  Int   iPicStride = pcCU->getPic()->getStride(compID);
  Bool  bNeighborFlags[4 * MAX_NUM_SPU_W + 1];
  Int   iNumIntraNeighbor = 0;

  bNeighborFlags[iLeftUnits] = isAboveLeftAvailable( pcCU, uiPartIdxLT );
  iNumIntraNeighbor += bNeighborFlags[iLeftUnits] ? 1 : 0;
  iNumIntraNeighbor  += isAboveAvailable     ( pcCU, uiPartIdxLT, uiPartIdxRT, (bNeighborFlags + iLeftUnits + 1)                    );
  iNumIntraNeighbor  += isAboveRightAvailable( pcCU, uiPartIdxLT, uiPartIdxRT, (bNeighborFlags + iLeftUnits + 1 + iTUWidthInUnits ) );
  iNumIntraNeighbor  += isLeftAvailable      ( pcCU, uiPartIdxLT, uiPartIdxLB, (bNeighborFlags + iLeftUnits - 1)                    );
  iNumIntraNeighbor  += isBelowLeftAvailable ( pcCU, uiPartIdxLT, uiPartIdxLB, (bNeighborFlags + iLeftUnits - 1 - iTUHeightInUnits) );

  bAbove = true;
  bLeft  = true;

  const ChromaFormat chFmt       = rTu.GetChromaFormat();
  const UInt         uiROIWidth  = uiTuWidth2+1;
  const UInt         uiROIHeight = uiTuHeight2+1;

  assert(uiROIWidth*uiROIHeight <= m_iYuvExtSize);

#ifdef DEBUG_STRING
  std::stringstream ss(stringstream::out);
#endif

  {
    Pel *piAdiTemp   = m_piYuvExt[compID][PRED_BUF_UNFILTERED];
    Pel *piRoiOrigin = pcCU->getPic()->getPicYuvRec()->getAddr(compID, pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiZorderIdxInPart);
    fillReferenceSamples (g_bitDepth[chType], pcCU, piRoiOrigin, piAdiTemp, bNeighborFlags, iNumIntraNeighbor,  iUnitWidth, iUnitHeight, iAboveUnits, iLeftUnits,
                          uiTuWidth, uiTuHeight, uiROIWidth, uiROIHeight, iPicStride, toChannelType(compID), chFmt);


#ifdef DEBUG_STRING
    if (DebugOptionList::DebugString_Pred.getInt()&DebugStringGetPredModeMask(MODE_INTRA))
    {
      ss << "###: generating Ref Samples for channel " << compID << " and " << rTu.getRect(compID).width << " x " << rTu.getRect(compID).height << "\n";
      for (UInt y=0; y<uiROIHeight; y++)
      {
        ss << "###: - ";
        for (UInt x=0; x<uiROIWidth; x++)
        {
          if (x==0 || y==0)
            ss << piAdiTemp[y*uiROIWidth + x] << ", ";
//          if (x%16==15) ss << "\nPart size: ~ ";
        }
        ss << "\n";
      }
    }
#endif

    if (bFilterRefSamples)
    {
      // generate filtered intra prediction samples

            Int          stride    = uiROIWidth;
      const Pel         *piSrcPtr  = piAdiTemp                             + (stride * uiTuHeight2); // bottom left
            Pel         *piDestPtr = m_piYuvExt[compID][PRED_BUF_FILTERED] + (stride * uiTuHeight2); // bottom left

      //------------------------------------------------

      //TODO: RExt - investigate the use of this for chroma (luma only for now to match HM)
      Bool useStrongIntraSmoothing = isLuma(chType) && pcCU->getSlice()->getSPS()->getUseStrongIntraSmoothing();

      const Pel bottomLeft = piAdiTemp[stride * uiTuHeight2];
      const Pel topLeft    = piAdiTemp[0];
      const Pel topRight   = piAdiTemp[uiTuWidth2];

      if (useStrongIntraSmoothing)
      {
        const Int  threshold     = 1 << (g_bitDepth[chType] - 5);
        const Bool bilinearLeft  = abs((bottomLeft + topLeft ) - (2 * piAdiTemp[stride * uiTuHeight])) < threshold; //difference between the
        const Bool bilinearAbove = abs((topLeft    + topRight) - (2 * piAdiTemp[         uiTuWidth ])) < threshold; //ends and the middle
        if ((uiTuWidth < 32) || (!bilinearLeft) || (!bilinearAbove))
          useStrongIntraSmoothing = false;
      }

      *piDestPtr = *piSrcPtr; // bottom left is not filtered
      piDestPtr -= stride;
      piSrcPtr  -= stride;

      //------------------------------------------------

      //left column (bottom to top)

      if (useStrongIntraSmoothing)
      {
        const Int shift = g_aucConvertToBit[uiTuHeight] + 3; //log2(uiTuHeight2)

        for(UInt i=1; i<uiTuHeight2; i++, piDestPtr-=stride)
        {
          *piDestPtr = (((uiTuHeight2 - i) * bottomLeft) + (i * topLeft) + uiTuHeight) >> shift;
        }

        piSrcPtr -= stride * (uiTuHeight2 - 1);
      }
      else
      {
        for(UInt i=1; i<uiTuHeight2; i++, piDestPtr-=stride, piSrcPtr-=stride)
        {
          *piDestPtr = ( piSrcPtr[stride] + 2*piSrcPtr[0] + piSrcPtr[-stride] + 2 ) >> 2;
        }
      }

      //------------------------------------------------

      //top-left

      if (useStrongIntraSmoothing)
      {
        *piDestPtr = piSrcPtr[0];
      }
      else
      {
        *piDestPtr = ( piSrcPtr[stride] + 2*piSrcPtr[0] + piSrcPtr[1] + 2 ) >> 2;
      }
      piDestPtr += 1;
      piSrcPtr  += 1;

      //------------------------------------------------

      //top row (left-to-right)

      if (useStrongIntraSmoothing)
      {
        const Int shift = g_aucConvertToBit[uiTuWidth] + 3; //log2(uiTuWidth2)

        for(UInt i=1; i<uiTuWidth2; i++, piDestPtr++)
        {
          *piDestPtr = (((uiTuWidth2 - i) * topLeft) + (i * topRight) + uiTuWidth) >> shift;
        }

        piSrcPtr += uiTuWidth2 - 1;
      }
      else
      {
        for(UInt i=1; i<uiTuWidth2; i++, piDestPtr++, piSrcPtr++)
        {
          *piDestPtr = ( piSrcPtr[1] + 2*piSrcPtr[0] + piSrcPtr[-1] + 2 ) >> 2;
        }
      }

      //------------------------------------------------

      *piDestPtr=*piSrcPtr; // far right is not filtered

#ifdef DEBUG_STRING
    if (DebugOptionList::DebugString_Pred.getInt()&DebugStringGetPredModeMask(MODE_INTRA))
    {
      ss << "###: filtered result for channel " << compID <<"\n";
      for (UInt y=0; y<uiROIHeight; y++)
      {
        ss << "###: - ";
        for (UInt x=0; x<uiROIWidth; x++)
        {
          if (x==0 || y==0)
            ss << m_piYuvExt[compID][PRED_BUF_FILTERED][y*uiROIWidth + x] << ", ";
//          if (x%16==15) ss << "\nPart size: ~ ";
        }
        ss << "\n";
      }
    }
#endif


    }
  }
  DEBUG_STRING_APPEND(sDebug, ss.str())
}

Void fillReferenceSamples( const Int bitDepth, TComDataCU* pcCU, const Pel* piRoiOrigin, Pel* piAdiTemp, const Bool* bNeighborFlags,
                           const Int iNumIntraNeighbor, const Int unitWidth, const Int unitHeight, const Int iAboveUnits, const Int iLeftUnits,
                           const UInt uiCuWidth, const UInt uiCuHeight, const UInt uiWidth, const UInt uiHeight, const Int iPicStride,
                           const ChannelType chType, const ChromaFormat chFmt )
{
  // NOTE: RExt - This function has been modified for increased flexibility as part of the Square TU 4:2:2 implementation
  const Pel* piRoiTemp;
  Int  i, j;
  Int  iDCValue = 1 << (bitDepth - 1);
  const Int iTotalUnits = iAboveUnits + iLeftUnits + 1; //+1 for top-left

  if (iNumIntraNeighbor == 0)
  {
    // Fill border with DC value
    for (i=0; i<uiWidth; i++)
    {
      piAdiTemp[i] = iDCValue;
    }
    for (i=1; i<uiHeight; i++)
    {
      piAdiTemp[i*uiWidth] = iDCValue;
    }
  }
  else if (iNumIntraNeighbor == iTotalUnits)
  {
    // Fill top-left border and top and top right with rec. samples
    piRoiTemp = piRoiOrigin - iPicStride - 1;

    for (i=0; i<uiWidth; i++)
    {
      piAdiTemp[i] = piRoiTemp[i];
    }

    // Fill left and below left border with rec. samples
    piRoiTemp = piRoiOrigin - 1;

    for (i=1; i<uiHeight; i++)
    {
      piAdiTemp[i*uiWidth] = *(piRoiTemp);
      piRoiTemp += iPicStride;
    }
  }
  else // reference samples are partially available
  {
    // all above units have "unitWidth" samples each, all left/below-left units have "unitHeight" samples each
    const Int  iTotalSamples = (iLeftUnits * unitHeight) + ((iAboveUnits + 1) * unitWidth);
    Pel  piAdiLine[5 * MAX_CU_SIZE];
    Pel  *piAdiLineTemp;
    const Bool *pbNeighborFlags;


    // Initialize
    for (i=0; i<iTotalSamples; i++)
    {
      piAdiLine[i] = iDCValue;
    }

    // Fill top-left sample
    piRoiTemp = piRoiOrigin - iPicStride - 1;
    piAdiLineTemp = piAdiLine + (iLeftUnits * unitHeight);
    pbNeighborFlags = bNeighborFlags + iLeftUnits;
    if (*pbNeighborFlags)
    {
      Pel topLeftVal=piRoiTemp[0];
      for (i=0; i<unitWidth; i++)
      {
        piAdiLineTemp[i] = topLeftVal;
      }
    }

    // Fill left & below-left samples (downwards)
    piRoiTemp += iPicStride;
    piAdiLineTemp--;
    pbNeighborFlags--;

    for (j=0; j<iLeftUnits; j++)
    {
      if (*pbNeighborFlags)
      {
        for (i=0; i<unitHeight; i++)
        {
          piAdiLineTemp[-i] = piRoiTemp[i*iPicStride];
        }
      }
      piRoiTemp += unitHeight*iPicStride;
      piAdiLineTemp -= unitHeight;
      pbNeighborFlags--;
    }

    // Fill above & above-right samples (left-to-right) (each unit has "unitWidth" samples)
    piRoiTemp = piRoiOrigin - iPicStride;
    // offset line buffer by iNumUints2*unitHeight (for left/below-left) + unitWidth (for above-left)
    piAdiLineTemp = piAdiLine + (iLeftUnits * unitHeight) + unitWidth;
    pbNeighborFlags = bNeighborFlags + iLeftUnits + 1;
    for (j=0; j<iAboveUnits; j++)
    {
      if (*pbNeighborFlags)
      {
        for (i=0; i<unitWidth; i++)
        {
          piAdiLineTemp[i] = piRoiTemp[i];
        }
      }
      piRoiTemp += unitWidth;
      piAdiLineTemp += unitWidth;
      pbNeighborFlags++;
    }

    // Pad reference samples when necessary
    Int iCurrJnit = 0;
    Pel  *piAdiLineCur   = piAdiLine;
    const UInt piAdiLineTopRowOffset = iLeftUnits * (unitHeight - unitWidth);

    if (!bNeighborFlags[0])
    {
      // very bottom unit of bottom-left; at least one unit will be valid.
      {
        Int   iNext = 1;
        while (iNext < iTotalUnits && !bNeighborFlags[iNext])
        {
          iNext++;
        }
        Pel *piAdiLineNext = piAdiLine + ((iNext < iLeftUnits) ? (iNext * unitHeight) : (piAdiLineTopRowOffset + (iNext * unitWidth)));
        const Pel refSample = *piAdiLineNext;
        // Pad unavailable samples with new value
        Int iNextOrTop = std::min<Int>(iNext, iLeftUnits);
        // fill left column
        while (iCurrJnit < iNextOrTop)
        {
          for (i=0; i<unitHeight; i++)
          {
            piAdiLineCur[i] = refSample;
          }
          piAdiLineCur += unitHeight;
          iCurrJnit++;
        }
        // fill top row
        while (iCurrJnit < iNext)
        {
          for (i=0; i<unitWidth; i++)
          {
            piAdiLineCur[i] = refSample;
          }
          piAdiLineCur += unitWidth;
          iCurrJnit++;
        }
      }
    }

    // pad all other reference samples.
    while (iCurrJnit < iTotalUnits)
    {
      if (!bNeighborFlags[iCurrJnit]) // samples not available
      {
        {
          const Int numSamplesInCurrUnit = (iCurrJnit >= iLeftUnits) ? unitWidth : unitHeight;
          const Pel refSample = *(piAdiLineCur-1);
          for (i=0; i<numSamplesInCurrUnit; i++)
          {
            piAdiLineCur[i] = refSample;
          }
          piAdiLineCur += numSamplesInCurrUnit;
          iCurrJnit++;
        }
      }
      else
      {
        piAdiLineCur += (iCurrJnit >= iLeftUnits) ? unitWidth : unitHeight;
        iCurrJnit++;
      }
    }

    // Copy processed samples

    piAdiLineTemp = piAdiLine + uiHeight + unitWidth - 2;
    // top left, top and top right samples
    for (i=0; i<uiWidth; i++)
    {
      piAdiTemp[i] = piAdiLineTemp[i];
    }

    piAdiLineTemp = piAdiLine + uiHeight - 1;
    for (i=1; i<uiHeight; i++)
    {
      piAdiTemp[i*uiWidth] = piAdiLineTemp[-i];
    }
  }
}

/** Get pointer to reference samples for intra prediction
 * \param uiDirMode   prediction mode index
 * \param log2BlkSize size of block (2 = 4x4, 3 = 8x8, 4 = 16x16, 5 = 32x32, 6 = 64x64)
 * \param piAdiBuf    pointer to unfiltered reference samples
 * \return            pointer to (possibly filtered) reference samples
 *
 * The prediction mode index is used to determine whether a smoothed reference sample buffer is returned.
 */

Bool TComPrediction::filteringIntraReferenceSamples(const ComponentID compID, UInt uiDirMode, UInt uiTuChWidth, UInt uiTuChHeight, const ChromaFormat chFmt, const Bool intraReferenceSmoothingDisabled)
{
  Bool bFilter;

  if (!filterIntraReferenceSamples(toChannelType(compID), chFmt, intraReferenceSmoothingDisabled))
  {
    bFilter=false;
  }
  else
  {
    assert(uiTuChWidth>=4 && uiTuChHeight>=4 && uiTuChWidth<128 && uiTuChHeight<128);

    if (uiDirMode == DC_IDX)
    {
      bFilter=false; //no smoothing for DC or LM chroma
    }
    else
    {
      Int diff = min<Int>(abs((Int) uiDirMode - HOR_IDX), abs((Int)uiDirMode - VER_IDX));
      UInt sizeIndex=g_aucConvertToBit[uiTuChWidth];
      assert(sizeIndex < MAX_INTRA_FILTER_DEPTHS);
      bFilter = diff > m_aucIntraFilter[toChannelType(compID)][sizeIndex];
    }
  }
  return bFilter;
}

Bool isAboveLeftAvailable( TComDataCU* pcCU, UInt uiPartIdxLT )
{
  Bool bAboveLeftFlag;
  UInt uiPartAboveLeft;
  TComDataCU* pcCUAboveLeft = pcCU->getPUAboveLeft( uiPartAboveLeft, uiPartIdxLT );
  if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
  {
    bAboveLeftFlag = ( pcCUAboveLeft && pcCUAboveLeft->isConstrainedIntra( uiPartAboveLeft ) );
  }
  else
  {
    bAboveLeftFlag = (pcCUAboveLeft ? true : false);
  }
  return bAboveLeftFlag;
}

Int isAboveAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool *bValidFlags )
{
  const UInt uiRasterPartBegin = g_auiZscanToRaster[uiPartIdxLT];
  const UInt uiRasterPartEnd = g_auiZscanToRaster[uiPartIdxRT]+1;
  const UInt uiIdxStep = 1;
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiRasterPart = uiRasterPartBegin; uiRasterPart < uiRasterPartEnd; uiRasterPart += uiIdxStep )
  {
    UInt uiPartAbove;
    TComDataCU* pcCUAbove = pcCU->getPUAbove( uiPartAbove, g_auiRasterToZscan[uiRasterPart] );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCUAbove && pcCUAbove->isConstrainedIntra( uiPartAbove ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if (pcCUAbove)
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags++;
  }
  return iNumIntra;
}

Int isLeftAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool *bValidFlags )
{
  const UInt uiRasterPartBegin = g_auiZscanToRaster[uiPartIdxLT];
  const UInt uiRasterPartEnd = g_auiZscanToRaster[uiPartIdxLB]+1;
  const UInt uiIdxStep = pcCU->getPic()->getNumPartInWidth();
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiRasterPart = uiRasterPartBegin; uiRasterPart < uiRasterPartEnd; uiRasterPart += uiIdxStep )
  {
    UInt uiPartLeft;
    TComDataCU* pcCULeft = pcCU->getPULeft( uiPartLeft, g_auiRasterToZscan[uiRasterPart] );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCULeft && pcCULeft->isConstrainedIntra( uiPartLeft ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if ( pcCULeft )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags--; // opposite direction
  }

  return iNumIntra;
}

Int isAboveRightAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool *bValidFlags )
{
  const UInt uiNumUnitsInPU = g_auiZscanToRaster[uiPartIdxRT] - g_auiZscanToRaster[uiPartIdxLT] + 1;
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiOffset = 1; uiOffset <= uiNumUnitsInPU; uiOffset++ )
  {
    UInt uiPartAboveRight;
    TComDataCU* pcCUAboveRight = pcCU->getPUAboveRightAdi( uiPartAboveRight, uiPartIdxRT, uiOffset );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCUAboveRight && pcCUAboveRight->isConstrainedIntra( uiPartAboveRight ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if ( pcCUAboveRight )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags++;
  }

  return iNumIntra;
}

Int isBelowLeftAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool *bValidFlags )
{
  const UInt uiNumUnitsInPU = (g_auiZscanToRaster[uiPartIdxLB] - g_auiZscanToRaster[uiPartIdxLT]) / pcCU->getPic()->getNumPartInWidth() + 1;
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiOffset = 1; uiOffset <= uiNumUnitsInPU; uiOffset++ )
  {
    UInt uiPartBelowLeft;
    TComDataCU* pcCUBelowLeft = pcCU->getPUBelowLeftAdi( uiPartBelowLeft, uiPartIdxLB, uiOffset );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCUBelowLeft && pcCUBelowLeft->isConstrainedIntra( uiPartBelowLeft ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if ( pcCUBelowLeft )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags--; // opposite direction
  }

  return iNumIntra;
}

#if PALETTE_MODE
Void TComPrediction::deriveRunLuma (TComDataCU* pcCU, Pel* pOrg,  Pel *pPalette,  Pel* pValue, Pel* pSPoint, Pel* pEscapeFlag, Pel* pPixelPredFlag,   Pel * paRecoValue, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{
  UInt uiTotal = uiHeight * uiWidth, uiIdx = 0, uiStride = uiWidth;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;
  Int iTemp = 0;

  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;

    uiRun = 0;
    Bool RunValid = CalRunLuma     ( pcCU, pOrg, pPalette,   pValue, pSPoint, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiRun);

    uiCopyRun = 0;
    Bool CopyValid = CalCopyLuma    ( pcCU, pOrg, pPalette,   pValue, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiCopyRun);

    if (CopyValid == 0 && RunValid == 0)
    {
      pEscapeFlag[uiIdx] = 1;
      pPixelPredFlag[uiIdx] = CalcPixelPredLuma(pcCU, pOrg, pPalette, pValue, pEscapeFlag,   paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiStartPos );
      pSPoint[uiIdx] = 0;
      uiIdx++;
    }
    else
    {
      pEscapeFlag[uiIdx] = 0;
      pPixelPredFlag[uiIdx] = 0;
      paRecoValue [uiIdx] = pPalette[pValue[uiIdx]];
      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiIdx] = 1;
        pRun[uiIdx]  = uiCopyRun-1;

        uiIdx++;
        iTemp = uiCopyRun - 1;
        while (iTemp>0)
        {
          pValue[uiIdx] = pValue[uiIdx - uiStride];
          paRecoValue[uiIdx] = pPalette[pValue[uiIdx]];
          pSPoint[uiIdx] = 1;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
      else
      {
        pSPoint[uiIdx] = 0;
        pRun[uiIdx] = uiRun;
        uiIdx++;
        iTemp = uiRun;
        while (iTemp>0)
        {
          pValue[uiIdx] = pValue[uiIdx - 1];
          paRecoValue [uiIdx] = pPalette[pValue[uiIdx]];
          pSPoint[uiIdx] = 0;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
    }
  }
  assert (uiIdx == uiTotal);
}

Void  TComPrediction:: deriveRunChroma (TComDataCU* pcCU, Pel* pOrg [2],  Pel *pPalette [2],  Pel* pValue[2], Pel* pSPoint, Pel* pEscapeFlag, Pel* pPixelPredFlag,   Pel * paRecoValue[2], Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{

  UInt uiTotal = uiHeight * uiWidth, uiIdx = 0;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;

  Int iTemp = 0;
  UInt uiStride = uiWidth;
  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;
    uiRun = 0;

    Bool RunValid = false;
    RunValid = CalRunChroma     ( pcCU, pOrg, pPalette,   pValue[0], pSPoint, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiRun);
 
    uiCopyRun = 0;
    Bool CopyValid = false;
    CopyValid =   CalCopyChroma   ( pcCU, pOrg, pPalette,   pValue[0], pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiCopyRun);
 
    if (CopyValid == 0 && RunValid == 0)
    {
      pEscapeFlag[uiIdx] = 1;
      pPixelPredFlag[uiIdx] = CalcPixelPredChroma(pcCU, pOrg, pPalette, pValue, pEscapeFlag,   paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiStartPos );
      pSPoint[uiIdx] = 0;
      uiIdx++;
    }
    else
    {
      pEscapeFlag[uiIdx] = 0;
      pPixelPredFlag[uiIdx] = 0;
      paRecoValue[0] [uiIdx] = pPalette[0][pValue[0][uiIdx]];
      paRecoValue[1] [uiIdx] = pPalette[1][pValue[0][uiIdx]];

      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiIdx] = 1;
        pRun[uiIdx]  = uiCopyRun-1;
        uiIdx++;
        iTemp = uiCopyRun - 1;
        while (iTemp>0)
        {
          pValue[0][uiIdx] = pValue[0][uiIdx-uiStride];
          paRecoValue[0] [uiIdx] = pPalette[0][pValue[0][uiIdx]];
          paRecoValue[1] [uiIdx] = pPalette[1][pValue[0][uiIdx]];

          pSPoint[uiIdx] = 1;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
      else
      {
        pSPoint[uiIdx] = 0;
        pRun[uiIdx]  = uiRun;
        uiIdx++;
        iTemp = uiRun;
        while (iTemp>0)
        {
          pValue[0][uiIdx] = pValue[0][uiIdx-1];
          paRecoValue[0] [uiIdx] = pPalette[0][pValue[0][uiIdx]];
          paRecoValue[1] [uiIdx] = pPalette[1][pValue[0][uiIdx]];

          pSPoint[uiIdx] = 0;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
    }
  }

  assert (uiIdx == uiTotal);
}


Void TComPrediction::deriveRun(TComDataCU* pcCU, Pel* pOrg [3],  Pel *pPalette [3],  Pel* pValue, Pel* pSPoint, Pel* pEscapeFlag, Pel* pPixelPredFlag,  Pel** paPixelValue, Pel ** paRecoValue, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{
  UInt   uiTotal = uiHeight * uiWidth, uiIdx = 0;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;

  Int iTemp = 0;
  UInt uiStride = uiWidth;
  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;
    uiRun = 0;

    Bool RunValid = false;
    RunValid = CalRun     ( pcCU, pOrg, pPalette,   pValue, pSPoint, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiRun);
    uiCopyRun = 0;
    Bool CopyValid = false;
    CopyValid =   CalCopy    ( pcCU, pOrg, pPalette,   pValue, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiCopyRun);

    if (CopyValid == 0 && RunValid == 0)
    {
      pEscapeFlag[uiIdx] = 1;
      pPixelPredFlag[uiIdx] = CalcPixelPred(pcCU, pOrg, pPalette, pValue, pEscapeFlag,  paPixelValue, paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiStartPos );

      pSPoint[uiIdx] = 0;
      uiIdx++;
    }
    else
    {
      pEscapeFlag[uiIdx] = 0;
      pPixelPredFlag[uiIdx] = 0;

      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiIdx] = 1;
        pRun[uiIdx]  = uiCopyRun-1;
        uiIdx++;
        iTemp = uiCopyRun - 1;
        while (iTemp>0)
        {
          pValue[uiIdx] = pValue[uiIdx-uiStride];
          pSPoint[uiIdx] = 1;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
      else
      {
        pSPoint[uiIdx] = 0;
        pRun[uiIdx] = uiRun;

        uiIdx++;
        iTemp = uiRun;
        while (iTemp > 0)
        {
          pValue[uiIdx] = pValue[uiIdx - 1];
          pSPoint[uiIdx] = 0;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
    }
  }
  assert (uiIdx == uiTotal);
}
#if PLT_IMPROVE_GEN
#if PLT_SIZE_ENC_IDX_REFINE
Void TComPrediction::derivePLTCLuma(TComDataCU* pcCU, Pel *Palette, Pel* pSrc, UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  UInt uiTotalSize = uiHeight * uiWidth;
  SortingElementLuma *psList = new SortingElementLuma[uiTotalSize];
  SortingElementLuma sElement;
  SortingElementLuma *pListSort = new SortingElementLuma[uiTotalSize + 1];

  UInt uiPos;
  UInt curPLTSize = 0;
  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiStride + uiX;
      if (m_cIndexBlock[uiPos] >= 0)
      {
        curPLTSize = m_cIndexBlock[uiPos] > curPLTSize ? m_cIndexBlock[uiPos] : curPLTSize;
        UInt uiPLTIDX = m_cIndexBlock[uiPos];
        psList[uiPLTIDX].accumulateDataFrom(pSrc[uiPos]);
      }
    }
  }
  curPLTSize++;

  //bubble sorting, high frequency -> low frequency
  for (Int i = 0; i < curPLTSize; i++)
  {
    pListSort[i].copyAllFrom(psList[i]);
    for (Int j = i; j > 0; j--)
    {
      if (pListSort[j].uiCnt > pListSort[j - 1].uiCnt)
      {
        sElement.copyAllFrom(pListSort[j - 1]);
        pListSort[j - 1].copyAllFrom(pListSort[j]);
        pListSort[j].copyAllFrom(sElement);
      }
      else
      {
        break;
      }
    }
  }

  assert(uiPLTSize == curPLTSize);
  uiPLTSize = 0;
  for (Int i = 0; i < curPLTSize; i++)
  {
    Double Paletted = pListSort[i].uiDataLuma * 1.0 / pListSort[i].uiCnt;
    UInt uiMostSimilar = 0;
    Double dMinDistance = MAX_DOUBLE;
    for (UInt uiY = 0; uiY < uiHeight; uiY++)
    {
      for (UInt uiX = 0; uiX < uiWidth; uiX++)
      {
        uiPos = uiY * uiStride + uiX;
        Double dDist = abs(pSrc[uiPos] - Paletted);
        if (dDist < dMinDistance)
        {
          dMinDistance = dDist;
          uiMostSimilar = uiPos;
        }
      }
    }
    Palette[uiPLTSize] = pSrc[uiMostSimilar];
    uiPLTSize++;
  }

  delete[] psList;
  delete[] pListSort;
}
#endif
Void  TComPrediction::derivePLTLuma (TComDataCU* pcCU, Pel *Palette, Pel* pSrc,  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight * uiWidth;
  SortingElementLuma *psList = new SortingElementLuma [uiTotalSize];
  SortingElementLuma sElement;

  //UInt uiDictMaxSize = MAX_PLT_SIZE;
  //SortingElementLuma *pListSort = new SortingElementLuma [MAX_PLT_SIZE+1];
  SortingElementLuma *pListSort = new SortingElementLuma[uiTotalSize + 1];

  UInt uiIdx = 0;
  //UInt uiShift = 0;
  UInt uiPos;
  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    //uiShift     =  uiY * uiWidth;
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiWidth + uiX;
      //sElement.setAll (pSrc[uiShift + uiX]);
      sElement.setAll(pSrc[uiPos]);
      Int i = 0;
      for (i = uiIdx - 1; i >= 0; i--)
      {
        //if (psList[i].almostEqualData(sElement, iErrorLimit))
        if (psList[i].exactlyEqualData(sElement))
        {
          psList[i].uiCnt++;
          break;
        }
      }
      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);
        psList[uiIdx].uiCnt++;
        uiIdx++;
      }
    }
  }

  for (Int i = 0; i < uiIdx; i++)
  {
    pListSort[i].copyAllFrom (psList[i]);
    for (Int j = i; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j - 1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j - 1]) ;
        pListSort[j - 1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = 0;
  for (Int i = 0; i < uiIdx; i++)
  {
    for (Int j = i - 1; j >= 0; j--)
    {
      if (pListSort[j].uiCnt && pListSort[i].almostEqualData(pListSort[j], iErrorLimit))
      {
        pListSort[j].uiCnt += pListSort[i].uiCnt;
        pListSort[i].uiCnt = 0;
        break;
      }
    }

    if (pListSort[i].uiCnt)
    {
      Palette[uiPLTSize] = pListSort[i].uiDataLuma;
      uiPLTSize++;
      if (uiPLTSize == MAX_PLT_SIZE)
      { 
        break;
      }
    }
  }

  delete [] psList;
  delete [] pListSort;
}
#if PLT_SIZE_ENC_IDX_REFINE
Void TComPrediction::derivePLTCChroma(TComDataCU* pcCU, Pel *Palette[2], Pel* pSrc[2], UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  UInt uiTotalSize = uiHeight * uiWidth;
  SortingElementChroma *psList = new SortingElementChroma[uiTotalSize];
  SortingElementChroma sElement;
  SortingElementChroma *pListSort = new SortingElementChroma[uiTotalSize + 1];

  UInt uiPos;
  UInt curPLTSize = 0;
  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiStride + uiX;
      if (m_cIndexBlockChroma[uiPos] >= 0)
      {
        curPLTSize = m_cIndexBlockChroma[uiPos] > curPLTSize ? m_cIndexBlockChroma[uiPos] : curPLTSize;
        UInt uiPLTIDX = m_cIndexBlockChroma[uiPos];
        psList[uiPLTIDX].accumulateDataFrom(pSrc[0][uiPos], pSrc[1][uiPos]);
      }
    }
  }
  curPLTSize++;

  //bubble sorting, high frequency -> low frequency
  for (Int i = 0; i < curPLTSize; i++)
  {
    pListSort[i].copyAllFrom(psList[i]);
    for (Int j = i; j > 0; j--)
    {
      if (pListSort[j].uiCnt > pListSort[j - 1].uiCnt)
      {
        sElement.copyAllFrom(pListSort[j - 1]);
        pListSort[j - 1].copyAllFrom(pListSort[j]);
        pListSort[j].copyAllFrom(sElement);
      }
      else
      {
        break;
      }
    }
  }

  assert(uiPLTSize == curPLTSize);
  uiPLTSize = 0;
  for (Int i = 0; i < curPLTSize; i++)
  {
    Double PaletteD0 = pListSort[i].uiData[0] * 1.0 / pListSort[i].uiCnt;
    Double PaletteD1 = pListSort[i].uiData[1] * 1.0 / pListSort[i].uiCnt;
    UInt uiMostSimilar = 0;
    Double dMinDistance = MAX_DOUBLE;
    for (UInt uiY = 0; uiY < uiHeight; uiY++)
    {
      for (UInt uiX = 0; uiX < uiWidth; uiX++)
      {
        uiPos = uiY * uiStride + uiX;
        Double dDist = abs(pSrc[0][uiPos] - PaletteD0) + abs(pSrc[1][uiPos] - PaletteD1);
        if (dDist < dMinDistance)
        {
          dMinDistance = dDist;
          uiMostSimilar = uiPos;
        }
      }
    }
    Palette[0][uiPLTSize] = pSrc[0][uiMostSimilar];
    Palette[1][uiPLTSize] = pSrc[1][uiMostSimilar];
    uiPLTSize++;
  }

  delete[] psList;
  delete[] pListSort;
}
#endif
Void  TComPrediction::derivePLTChroma (TComDataCU* pcCU, Pel *Palette[2], Pel* pSrc[2],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{

  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0 : getPLTErrLimit();

  UInt uiTotalSize = uiHeight * uiWidth;
  SortingElementChroma *psList = new SortingElementChroma [uiTotalSize];
  SortingElementChroma sElement;

  //UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElementChroma *pListSort = new SortingElementChroma[uiTotalSize + 1];

  UInt uiIdx = 0;
  UInt uiPos;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      uiPos = uiY * uiWidth + uiX;
      sElement.setAll (pSrc[0][uiPos], pSrc[1][uiPos]);
      Int i = 0;
      for ( i = uiIdx - 1; i >= 0; i--)
      { 
        //if (psList[i].almostEqualData(sElement, iErrorLimit))
        if (psList[i].exactlyEqualData(sElement))
        {
          psList[i].uiCnt++;
          break;
        }
      }

      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);
        psList[uiIdx].uiCnt++;
        uiIdx++;
      }
    }
  }

  for (Int i = 0; i < uiIdx; i++)
  {
    pListSort[i].copyAllFrom (psList[i]);
    for (Int j = i; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j - 1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j - 1]) ;
        pListSort[j - 1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = 0;
  for (Int i = 0; i < uiIdx; i++)
  {
    for (Int j = i - 1; j >= 0; j--)
    {
      if (pListSort[j].uiCnt && pListSort[i].almostEqualData(pListSort[j], iErrorLimit))
      {
        pListSort[j].uiCnt += pListSort[i].uiCnt;
        pListSort[i].uiCnt = 0;
        break;
      }
    }

    if (pListSort[i].uiCnt)
    {
      Palette[0][uiPLTSize] = pListSort[i].uiData[0];
      Palette[1][uiPLTSize] = pListSort[i].uiData[1];
      uiPLTSize++;
      if (uiPLTSize == MAX_PLT_SIZE)
      {
        break;
      }
    }
  }

  delete [] psList;
  delete [] pListSort;
}

#else

Void  TComPrediction::derivePLTLuma (TComDataCU* pcCU, Pel *Palette, Pel* pSrc,  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{

  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElementLuma *psList = new SortingElementLuma [uiTotalSize];
  SortingElementLuma sElement;

  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElementLuma *pListSort = new SortingElementLuma [MAX_PLT_SIZE+1];

  UInt uiIdx = 0;
  UInt uiShift = 0;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    uiShift     =  uiY * uiWidth;
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      sElement.setAll (pSrc[uiShift + uiX]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      { 
        if (psList[i].almostEqualData(sElement, iErrorLimit))
        {

          psList[i].uiCnt++;

          break;
        }
      }

      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;
      }
    }
  }


  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0) ;
  }

  for (Int i = 0; i < uiTotalSize; i++)
  {
    pListSort[uiDictMaxSize].copyAllFrom (psList[i]);

    for (Int j = uiDictMaxSize; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = MAX_PLT_SIZE;
  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    Palette[i] = pListSort[i].uiDataLuma;
    if (pListSort[i].uiCnt == 0  )
    {
      uiPLTSize = i;
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;


}

Void  TComPrediction::derivePLTChroma (TComDataCU* pcCU, Pel *Palette[2], Pel* pSrc[2],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{

  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElementChroma *psList = new SortingElementChroma [uiTotalSize];
  SortingElementChroma sElement;

  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElementChroma *pListSort = new SortingElementChroma [MAX_PLT_SIZE+1];

  UInt uiIdx = 0;
  //UInt uiShift = 0, uiCbCrShift = 0;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    UInt uiShift     =  uiY * uiWidth;
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      sElement.setAll (pSrc[0][uiShift + uiX], pSrc[1][uiShift + uiX]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      { 
        if (psList[i].almostEqualData(sElement, iErrorLimit))
        {

          psList[i].uiCnt++;

          break;
        }
      }

      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;

      }
    }
  }


  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0, 0) ;
  }

  for (Int i = 0; i < uiTotalSize; i++)
  {
    pListSort[uiDictMaxSize].copyAllFrom (psList[i]);

    for (Int j = uiDictMaxSize; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = MAX_PLT_SIZE;
  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    Palette[0][i] = pListSort[i].uiData[0];
    Palette[1][i] = pListSort[i].uiData[1];
    if (pListSort[i].uiCnt == 0  )
    {
      uiPLTSize = i;
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;


}

#endif

Void   TComPrediction:: reorderPLTLuma (TComDataCU* pcCU, Pel *pPalette)
{
  UInt uiPLTSizePrev, uiDictMaxSize;
  Pel * pPalettePrev;
  Pel pPaletteTemp[MAX_PLT_SIZE];


  pPalettePrev = pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Y, uiPLTSizePrev);
  for (UInt i = 0; i < MAX_PLT_SIZE; i++)
  {
    pPaletteTemp[i] = pPalette[i];
  }


  pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Y, uiPLTSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(COMPONENT_Y, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
  Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_SIZE+1];

  memset (bReusedPrev, 0, MAX_PLT_SIZE+1);
  memset (bPredicted,  0, MAX_PLT_SIZE+1);


  UInt uiNumPLTRceived = uiDictMaxSize , uiNumPLTPredicted = 0;

  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
  {
    bReused = false;
    //Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
    {
      if (pPalettePrev[uiIdxPrev] == pPalette[uiIdxCurr] )
      {
        bReused = true;
        break;
      }

    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted [uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr] )
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted ++;
    }
  }
  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_SIZE; uiIdxPrev ++)
  {

    pcCU->setPrevPLTReusedFlagSubParts   ( COMPONENT_Y,  bReusedPrev[uiIdxPrev], uiIdxPrev, 0,  pcCU->getDepth(0)  );
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx ++)
  {
    if (bReusedPrev[uiPrevIdx])
    {

      pPalette[uiIdxCurr] = pPalettePrev[uiPrevIdx];
      uiIdxCurr ++;
    }
  }

  //UInt uiReceivedIdx = 0;
  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx ++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      pPalette[uiIdxCurr] = pPaletteTemp[uiIdx];
      uiIdxCurr++;
    }

  }
}
Void   TComPrediction:: reorderPLTChroma (TComDataCU* pcCU, Pel *pPalette[2])
{

  UInt uiPLTSizePrev, uiDictMaxSize;
  Pel * pPalettePrev[2];
  Pel pPaletteTemp[2][MAX_PLT_SIZE];

    for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
    {
      pPalettePrev[comp] = pcCU->getPLTPred (pcCU,  pcCU->getZorderIdxInCU(), comp+1, uiPLTSizePrev);
      for (UInt i = 0; i < MAX_PLT_SIZE; i++)
      {
        pPaletteTemp[comp][i] = pPalette[comp][i];
      }
    }
 
  pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Cb, uiPLTSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(COMPONENT_Cb, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
  Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_SIZE+1];

  memset (bReusedPrev, 0, MAX_PLT_SIZE+1);
  memset (bPredicted,  0, MAX_PLT_SIZE+1);


  UInt uiNumPLTRceived = uiDictMaxSize , uiNumPLTPredicted = 0;

  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
  {
    bReused = false;
    Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
    {
      iCounter = 0;

      for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
      {
        if (pPalettePrev[comp][uiIdxPrev] == pPalette[comp][uiIdxCurr] )
        {
          iCounter ++;
        }
      }
      if (iCounter == 2)
      {
        bReused = true;
        break;
      }
    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted [uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr] )
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted ++;
    }
  }
  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_SIZE; uiIdxPrev ++)
  {
    for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
    {
      pcCU->setPrevPLTReusedFlagSubParts   ( comp+1,  bReusedPrev[uiIdxPrev], uiIdxPrev, 0,  pcCU->getDepth(0)  );
    }
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx ++)
  {
    if (bReusedPrev[uiPrevIdx])
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPalettePrev[comp][uiPrevIdx];
      }
      uiIdxCurr ++;
    }
  }

  //UInt uiReceivedIdx = 0;
  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx ++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPaletteTemp[comp][uiIdx];
      }
      uiIdxCurr++;
    }

  }
}

#if PLT_IMPROVE_GEN
Void TComPrediction::preCalcPLTIndexLuma(TComDataCU* pcCU, Pel *Palette, Pel* pSrc, UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless ? 0 : getPLTErrLimit();
  UInt uiPos;
#if PLT_SIZE_ENC_IDX_REFINE
  Int iMap[MAX_PLT_SIZE] = { 0 };
#endif
  UInt uiBestIdx = 0;

  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiStride + uiX;
      UInt uiPLTIdx = 0;
      UInt uiMinError = MAX_UINT;

      while (uiPLTIdx < uiPLTSize)
      {
        UInt uiAbsError = abs(Palette[uiPLTIdx] - pSrc[uiPos]);

        if (uiAbsError < uiMinError)
        {
          uiBestIdx = uiPLTIdx;
          uiMinError = uiAbsError;

          if (uiMinError == 0)
          {
            break;
          }
        }
        uiPLTIdx++;
      }
      Pel oldIdx = m_cIndexBlockChroma[uiPos];
      m_cIndexBlock[uiPos] = uiBestIdx;
      if (oldIdx < 0)
      {
        m_cIndexBlock[uiPos] -= MAX_PLT_SIZE;
      }
#if PLT_SIZE_ENC_IDX_REFINE
      else
      {
        iMap[uiBestIdx]++;
      }
#endif
    }
  }

#if PLT_SIZE_ENC_IDX_REFINE
  while (true)
  {
    Int i = 0;
    while (i < uiPLTSize)
    {
      if (iMap[i] == 0)
      {
        uiPLTSize--;
        for (Int j = i; j < uiPLTSize; j++)
        {
          iMap[j] = iMap[j + 1];
          Palette[j] = Palette[j + 1];
        }
      }
      else
      {
        i++;
      }
    }

    uiBestIdx = 0;
    memset(iMap, 0, sizeof(Int)* MAX_PLT_SIZE);
    for (UInt uiY = 0; uiY < uiHeight; uiY++)
    {
      for (UInt uiX = 0; uiX < uiWidth; uiX++)
      {
        uiPos = uiY * uiStride + uiX;
        UInt uiPLTIdx = 0;
        UInt uiMinError = MAX_UINT;

        while (uiPLTIdx < uiPLTSize)
        {
          UInt uiAbsError = abs(Palette[uiPLTIdx] - pSrc[uiPos]);

          if (uiAbsError < uiMinError)
          {
            uiBestIdx = uiPLTIdx;
            uiMinError = uiAbsError;

            if (uiMinError == 0)
            {
              break;
            }
          }
          uiPLTIdx++;
        }
        m_cIndexBlock[uiPos] = uiBestIdx;
        if (uiMinError > iErrorLimit)
        {
          m_cIndexBlock[uiPos] -= MAX_PLT_SIZE;
        }
#if PLT_SIZE_ENC_IDX_REFINE
        else
        {
          iMap[uiBestIdx]++;
        }
#endif
      }
    }
    if (uiPLTSize == 0)
    {
      break;
    }
    for (Int i0 = 0; i0 < uiPLTSize; i0++)
    {
      if (iMap[i0] == 0)
      {
        continue;
      }
    }
    break;
  }
#endif
}

Void	TComPrediction::preCalcPLTIndexChroma(TComDataCU* pcCU, Pel *Palette[2], Pel* pSrc[2], UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless ? 0 : 2 * getPLTErrLimit();
  UInt uiPos;
#if PLT_SIZE_ENC_IDX_REFINE
  Int iMap[MAX_PLT_SIZE] = { 0 };
#endif
  UInt uiBestIdx = 0;

  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiStride + uiX;
      UInt uiPLTIdx = 0;
      UInt uiMinError = MAX_UINT;

      while (uiPLTIdx < uiPLTSize)
      {
        UInt uiAbsError = abs(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + abs(Palette[1][uiPLTIdx] - pSrc[1][uiPos]);

        
        if (uiAbsError < uiMinError)
        {
          uiBestIdx = uiPLTIdx;
          uiMinError = uiAbsError;

          if (uiMinError == 0)
          {
            break;
          }
        }
        uiPLTIdx++;
      }
      m_cIndexBlockChroma[uiPos] = uiBestIdx;
      if (uiMinError > iErrorLimit)
      {
        m_cIndexBlockChroma[uiPos] -= MAX_PLT_SIZE;
      }
#if PLT_SIZE_ENC_IDX_REFINE
      else
      {
        iMap[uiBestIdx]++;
      }
#endif
    }
  }
#if PLT_SIZE_ENC_IDX_REFINE
  while (true)
  {
    Int i1 = 0;
    while (i1 < uiPLTSize)
    {
      if (iMap[i1] == 0)
      {
        uiPLTSize--;
        for (Int j = i1; j < uiPLTSize; j++)
        {
          iMap[j] = iMap[j + 1];
          Palette[0][j] = Palette[0][j + 1];
          Palette[1][j] = Palette[1][j + 1];
        }
      }
      else
      {
        i1++;
      }
    }

    uiBestIdx = 0;
    memset(iMap, 0, MAX_PLT_SIZE * sizeof(Int));
    for (UInt uiY = 0; uiY < uiHeight; uiY++)
    {
      for (UInt uiX = 0; uiX < uiWidth; uiX++)
      {
        uiPos = uiY * uiStride + uiX;
        UInt uiPLTIdx = 0;
        UInt uiMinError = MAX_UINT;

        while (uiPLTIdx < uiPLTSize)
        {
          UInt uiAbsError = abs(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + abs(Palette[1][uiPLTIdx] - pSrc[1][uiPos]);
          if (uiAbsError < uiMinError)
          {
            uiBestIdx = uiPLTIdx;
            uiMinError = uiAbsError;

            if (uiMinError == 0)
            {
              break;
            }
          }
          uiPLTIdx++;
        }
        m_cIndexBlockChroma[uiPos] = uiBestIdx;
        if (uiMinError > iErrorLimit)
        {
          m_cIndexBlockChroma[uiPos] -= MAX_PLT_SIZE;
        }
#if PLT_SIZE_ENC_IDX_REFINE
        else
        {
          iMap[uiBestIdx]++;
        }
#endif
      }
    }
    if (uiPLTSize == 0)
    {
      break;
    }
    for (Int i = 0; i < uiPLTSize; i++)
    {
      if (iMap[i] == 0)
      {
        continue;
      }
    }
    break;
  }
#endif
}
#if PLT_SIZE_ENC_RD  
Void TComPrediction::preCalcPLTIndex(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt&uiPLTSize, TComRdCost* m_pcRdCost)
#else
Void TComPrediction::preCalcPLTIndex(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt uiPLTSize)
#endif
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless ? 0 : 3 * getPLTErrLimit();
  UInt uiPos;

  UInt uiBestIdx = 0; 
#if PLT_SIZE_ENC_RD
  UInt PLTClusterVol[MAX_PLT_SIZE];
  Int  PLTClusterSSE[MAX_PLT_SIZE];
  Int  PLTClusterQuantDist[MAX_PLT_SIZE];

  UInt uiPLTIndex = 0;
  UInt uiMaxBit[3] = { 0, 0, 0 };
  Pel  pOrg[3];
  Int  iQP[3];
  
  UInt uiShift = DISTORTION_PRECISION_ADJUSTMENT((g_bitDepth[0] - 8) << 1);
  UInt uiPLTSizeBit = 0, uiSize = MAX_PLT_SIZE;
  while (uiSize !=0 )
  {
    uiSize >>= 1;
    uiPLTSizeBit++;
  }
  uiPLTSizeBit--;

  for (UInt k = 0; k < MAX_PLT_SIZE; k++)
  {
    PLTClusterVol[k] = 0; //initialize the plt cluster counter as 0
    PLTClusterSSE[k] = 0; //initialize the plt cluster SSE as 0
    PLTClusterQuantDist[k] = 0; //initialize the plt clsuter quantization error as 0
  }

  pcCU->xCalcMaxBits(pcCU, uiMaxBit);
  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
  {
    QpParam cQP(*pcCU, ComponentID(ch));
    iQP[ch] = cQP.Qp;
  }
  
  
#endif

  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiWidth + uiX;
      UInt uiPLTIdx = 0;
      UInt uiMinError = MAX_UINT;
#if PLT_SIZE_ENC_RD
      UInt uiMinSSE = MAX_UINT;
#endif
      while (uiPLTIdx < uiPLTSize)
      {
        UInt uiAbsError = abs(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + abs(Palette[1][uiPLTIdx] - pSrc[1][uiPos]) + abs(Palette[2][uiPLTIdx] - pSrc[2][uiPos]);
#if PLT_SIZE_ENC_RD
        UInt uiSSE = (Palette[0][uiPLTIdx] - pSrc[0][uiPos]) *(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + (Palette[1][uiPLTIdx] - pSrc[1][uiPos])*(Palette[1][uiPLTIdx] - pSrc[1][uiPos]) + (Palette[2][uiPLTIdx] - pSrc[2][uiPos])*(Palette[2][uiPLTIdx] - pSrc[2][uiPos]);
        uiSSE >>= uiShift;
#endif
        if (uiAbsError < uiMinError)
        {
          uiBestIdx = uiPLTIdx;
          uiMinError = uiAbsError;
#if PLT_SIZE_ENC_RD
          uiMinSSE = uiSSE;
#endif
          if (uiMinError == 0)
          {
            break;
          }
        }
        uiPLTIdx++;
      }
      m_cIndexBlock[uiPos] = uiBestIdx;
      if (uiMinError > iErrorLimit)
      {
        m_cIndexBlock[uiPos] -= MAX_PLT_SIZE;
      }
#if PLT_SIZE_ENC_RD
     // fprintf(fp, "(i, j): %d\n", m_cIndexBlock[uiPos]);
      assert(m_cIndexBlock[uiPos] < MAX_PLT_SIZE);
      if (m_cIndexBlock[uiPos] >= 0) //non escape
      {
        PLTClusterVol[m_cIndexBlock[uiPos]]++;
        //calculate the cluster error
        PLTClusterSSE[m_cIndexBlock[uiPos]] += uiMinSSE;
        //calculate the quantization error
        pOrg[0] = pSrc[0][uiPos];
        pOrg[1] = pSrc[1][uiPos];
        pOrg[2] = pSrc[2][uiPos];
        PLTClusterQuantDist[m_cIndexBlock[uiPos]] += ((pcCU->CalcPixelQuantDist(pcCU, pOrg, iQP, uiMaxBit)) >> uiShift);
      }
#endif
    }
  }

#if PLT_SIZE_ENC_RD
 // fclose(fp);
  //given the current plt idx, check RD cost of this index when it is escape coded
  //motivation: for low occurance idx, it may be better for escape coded rather than index coded
  UInt uRemoveCounter = 0;
  UInt uiPLTSizeOrg = uiPLTSize;
  for (uiPLTIndex = 0; uiPLTIndex < uiPLTSizeOrg; uiPLTIndex++)
  {
    Int uUpdatedPLTIdx = Int(uiPLTIndex - uRemoveCounter);
   
    //Step 1: estimate the rate to code the index
    UInt uRateIdxCoded = (g_bitDepth[0] + 2 * g_bitDepth[1]) + PLTClusterVol[uiPLTIndex] * uiPLTSizeBit; //worst case
    //Step 2: estimate the rate to code the escape
    UInt uRateEscape = PLTClusterVol[uiPLTIndex] * (uiMaxBit[0] + uiMaxBit[1] + uiMaxBit[2]);
    //Step 3: calculate the cluster distortion for index coding 
    Int uDistCluster = PLTClusterSSE[uiPLTIndex];
    //Step 4: calculate the escape quantization error 
    // PLTClusterQuantDist[uiPLTIndex] ->available
    //Step 5: calculate the RD for index coding
    Double dIndexCoding = m_pcRdCost->calcRdCost(uRateIdxCoded, uDistCluster);
    //Step 6: calculate the RD for escape coding
    Double dEscapeCoding = m_pcRdCost->calcRdCost(uRateEscape, PLTClusterQuantDist[uiPLTIndex]);

    if (dEscapeCoding < dIndexCoding)
    {
      //remove this index entry 
      for (Int idx = uUpdatedPLTIdx + 1; idx < uiPLTSize; idx++)
      {
        Palette[0][idx - 1] = Palette[0][idx];
        Palette[1][idx - 1] = Palette[1][idx];
        Palette[2][idx - 1] = Palette[2][idx];
      }

      //reduce the plt size by one
      uiPLTSize--;
      uRemoveCounter++;

      //assign those pixels with escape mode
      for (UInt uiY = 0; uiY < uiHeight; uiY++)
        for (UInt uiX = 0; uiX < uiWidth; uiX++)
        {
          uiPos = uiY * uiWidth + uiX;
          if (m_cIndexBlock[uiPos] == uUpdatedPLTIdx)
          {
            m_cIndexBlock[uiPos] = 0 - MAX_PLT_SIZE; //set negative to indicate escape
          }
          else if (m_cIndexBlock[uiPos] > uUpdatedPLTIdx)
          {
            m_cIndexBlock[uiPos]--;
          }
        }

    }
  }
  
#endif
}
#endif


#if PLT_SIZE_ENC_RD
#if PLT_SIZE_ENC_IDX_REFINE
Void TComPrediction::CalcPLTIndex(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt &uiPLTSize)
#else
Void TComPrediction::CalcPLTIndex(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt uiPLTSize)
#endif
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless ? 0 : 3 * getPLTErrLimit();
  UInt uiPos;
#if PLT_SIZE_ENC_IDX_REFINE
  Int iMap[MAX_PLT_SIZE] = { 0 };
#endif
  UInt uiBestIdx = 0;

  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiWidth + uiX;
      UInt uiPLTIdx = 0;
      UInt uiMinError = MAX_UINT;

      while (uiPLTIdx < uiPLTSize)
      {
        UInt uiAbsError = abs(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + abs(Palette[1][uiPLTIdx] - pSrc[1][uiPos]) + abs(Palette[2][uiPLTIdx] - pSrc[2][uiPos]);

        if (uiAbsError < uiMinError)
        {
          uiBestIdx = uiPLTIdx;
          uiMinError = uiAbsError;

          if (uiMinError == 0)
          {
            break;
          }
        }
        uiPLTIdx++;
      }
      m_cIndexBlock[uiPos] = uiBestIdx;
      if (uiMinError > iErrorLimit)
      {
        m_cIndexBlock[uiPos] -= MAX_PLT_SIZE;
      }
#if PLT_SIZE_ENC_IDX_REFINE
      else
      {
        iMap[uiBestIdx]++;
      }
#endif
    }
  }
#if PLT_SIZE_ENC_IDX_REFINE
  while (true)
  {
    Int i = 0;
    while (i < uiPLTSize)
    {
      if (iMap[i] == 0)
      {
        uiPLTSize--;
        for (Int j = i; j < uiPLTSize; j++)
        {
          iMap[j] = iMap[j + 1];
          Palette[0][j] = Palette[0][j + 1];
          Palette[1][j] = Palette[1][j + 1];
          Palette[2][j] = Palette[2][j + 1];
        }
      }
      else
      {
        i++;
      }
    }

    uiBestIdx = 0;
    memset(iMap, 0, sizeof(Int)* MAX_PLT_SIZE);
    for (UInt uiY = 0; uiY < uiHeight; uiY++)
    {
      for (UInt uiX = 0; uiX < uiWidth; uiX++)
      {
        uiPos = uiY * uiWidth + uiX;
        UInt uiPLTIdx = 0;
        UInt uiMinError = MAX_UINT;

        while (uiPLTIdx < uiPLTSize)
        {
          UInt uiAbsError = abs(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + abs(Palette[1][uiPLTIdx] - pSrc[1][uiPos]) + abs(Palette[2][uiPLTIdx] - pSrc[2][uiPos]);

          if (uiAbsError < uiMinError)
          {
            uiBestIdx = uiPLTIdx;
            uiMinError = uiAbsError;

            if (uiMinError == 0)
            {
              break;
            }
          }
          uiPLTIdx++;
        }
        m_cIndexBlock[uiPos] = uiBestIdx;
        if (uiMinError > iErrorLimit)
        {
          m_cIndexBlock[uiPos] -= MAX_PLT_SIZE;
        }
        else
        {
          iMap[uiBestIdx]++;
        }
      }
    }
    if (uiPLTSize == 0)
    {
      break;
    }
    for (Int i0 = 0; i0 < uiPLTSize; i0++)
    {
      if (iMap[i0] == 0)
      {
        continue;
      }
    }
    break;
  }
#endif
}
#endif

Void  TComPrediction::reorderPLT (TComDataCU* pcCU, Pel *pPalette[3])
{

    UInt uiPLTSizePrev, uiDictMaxSize;
    Pel * pPalettePrev[3];
    Pel pPaletteTemp[3][MAX_PLT_SIZE];


    for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
    {
      pPalettePrev[comp] = pcCU->getPLTPred (pcCU,  pcCU->getZorderIdxInCU(), comp, uiPLTSizePrev);
      for (UInt i = 0; i < MAX_PLT_SIZE; i++)
      {
        pPaletteTemp[comp][i] = pPalette[comp][i];
      }
    }
 
  pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Y, uiPLTSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(COMPONENT_Y, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
  Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_SIZE+1];

  memset (bReusedPrev, 0, MAX_PLT_SIZE+1);
  memset (bPredicted,  0, MAX_PLT_SIZE+1);




  UInt uiNumPLTRceived = uiDictMaxSize , uiNumPLTPredicted = 0;

  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
  {
    bReused = false;
    Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
    {
      iCounter = 0;

      for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
      {
        if (pPalettePrev[comp][uiIdxPrev] == pPalette[comp][uiIdxCurr] )
        {
          iCounter ++;
        }
      }
      if (iCounter == 3)
      {
        bReused = true;
        break;
      }
    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted [uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr] )
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted ++;
    }
  }
  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_SIZE; uiIdxPrev ++)
  {
    for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
    {
      pcCU->setPrevPLTReusedFlagSubParts   ( comp,  bReusedPrev[uiIdxPrev], uiIdxPrev, 0,  pcCU->getDepth(0)  );
    }
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx ++)
  {
    if (bReusedPrev[uiPrevIdx])
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPalettePrev[comp][uiPrevIdx];
      }
      uiIdxCurr ++;
    }
  }

  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx ++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPaletteTemp[comp][uiIdx];
      }
      uiIdxCurr++;
    }

  }
}

#if PLT_IMPROVE_GEN
#if PLT_SIZE_ENC_IDX_REFINE
Void TComPrediction::derivePLTC(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  UInt uiTotalSize = uiHeight * uiWidth;
  SortingElement *psList = new SortingElement[uiTotalSize];
  SortingElement sElement;
  SortingElement *pListSort = new SortingElement[uiTotalSize + 1];

  UInt uiPos;
  UInt curPLTSize = 0;
  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiWidth + uiX;
      if (m_cIndexBlock[uiPos] >= 0)
      {
        curPLTSize = m_cIndexBlock[uiPos] > curPLTSize ? m_cIndexBlock[uiPos] : curPLTSize;
        UInt uiPLTIDX = m_cIndexBlock[uiPos];
        psList[uiPLTIDX].accumulateDataFrom(pSrc[0][uiPos], pSrc[1][uiPos], pSrc[2][uiPos]);
      }
    }
  }
  curPLTSize++;
  assert(uiPLTSize == curPLTSize);
  //bubble sorting, high frequency -> low frequency
  for (Int i = 0; i < curPLTSize; i++)
  {
    pListSort[i].copyAllFrom(psList[i]);
    for (Int j = i; j > 0; j--)
    {
      if (pListSort[j].uiCnt > pListSort[j - 1].uiCnt)
      {
        sElement.copyAllFrom(pListSort[j - 1]);
        pListSort[j - 1].copyAllFrom(pListSort[j]);
        pListSort[j].copyAllFrom(sElement);
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = 0;
  for (Int i = 0; i < curPLTSize; i++)
  {
    Double PaletteD0 = pListSort[i].uiData[0] * 1.0 / pListSort[i].uiCnt;
    Double PaletteD1 = pListSort[i].uiData[1] * 1.0 / pListSort[i].uiCnt;
    Double PaletteD2 = pListSort[i].uiData[2] * 1.0 / pListSort[i].uiCnt;
    UInt uiMostSimilar = 0;
    Double dMinDistance = MAX_DOUBLE;
    for (UInt uiY = 0; uiY < uiHeight; uiY++)
    {
      for (UInt uiX = 0; uiX < uiWidth; uiX++)
      {
        uiPos = uiY * uiWidth + uiX;
        Double dDist = abs(pSrc[0][uiPos] - PaletteD0) + abs(pSrc[1][uiPos] - PaletteD1) + abs(pSrc[2][uiPos] - PaletteD2);
        if (dDist < dMinDistance)
        {
          dMinDistance = dDist;
          uiMostSimilar = uiPos;
        }
      }
    }
    Palette[0][uiPLTSize] = pSrc[0][uiMostSimilar];
    Palette[1][uiPLTSize] = pSrc[1][uiMostSimilar];
    Palette[2][uiPLTSize] = pSrc[2][uiMostSimilar];
    uiPLTSize++;
  }

  delete[] psList;
  delete[] pListSort;
}
#endif
Void TComPrediction::derivePLT(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless ? 0 : getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElement *psList = new SortingElement [uiTotalSize];
  SortingElement sElement;
  SortingElement *pListSort = new SortingElement[uiTotalSize + 1];

  UInt uiIdx = 0;
  UInt uiPos;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      uiPos = uiY * uiWidth+ uiX;
      sElement.setAll (pSrc[0][uiPos], pSrc[1][uiPos], pSrc[2][uiPos]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      {
        if (psList[i].exactlyEqualData(sElement))
        {
          psList[i].uiCnt++;
          break;
        }
      }
      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;

      }
    }
  }

  //bubble sorting, high frequency -> low frequency
  for (Int i = 0; i < uiIdx; i++)
  {
    pListSort[i].copyAllFrom(psList[i]);
    for (Int j = i; j > 0; j--)
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = 0;
  for (Int i = 0; i < uiIdx; i++)
  {
    for (Int j = i - 1; j >= 0; j--)
    {
      if (pListSort[j].uiCnt && pListSort[i].almostEqualData(pListSort[j], iErrorLimit))
      {
        pListSort[j].uiCnt += pListSort[i].uiCnt;
        pListSort[i].uiCnt = 0;
        break;
      }
    }

    if (pListSort[i].uiCnt)
    {
      Palette[0][uiPLTSize] = pListSort[i].uiData[0];
      Palette[1][uiPLTSize] = pListSort[i].uiData[1];
      Palette[2][uiPLTSize] = pListSort[i].uiData[2];
      uiPLTSize++;
      if (uiPLTSize == MAX_PLT_SIZE)
      {
        break;
      }
    }
  }

  delete [] psList;
  delete [] pListSort;
}
#else
Void  TComPrediction::derivePLT (TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElement *psList = new SortingElement [uiTotalSize];
  SortingElement sElement;
  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElement *pListSort = new SortingElement [MAX_PLT_SIZE+1];
  UInt uiIdx = 0;
  UInt uiPos;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      uiPos = uiY * uiWidth+ uiX;
      sElement.setAll (pSrc[0][uiPos], pSrc[1][uiPos], pSrc[2][uiPos]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      { 
        if (psList[i].almostEqualData(sElement, iErrorLimit))
        {

          psList[i].uiCnt++;

          break;
        }
      }
      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;

      }
    }
  }

  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0, 0, 0) ;
  }

  //bubble sorting
  for (Int i = 0; i < uiTotalSize; i++)
  {
    pListSort[uiDictMaxSize].copyAllFrom (psList[i]);
    for (Int j = uiDictMaxSize; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = MAX_PLT_SIZE;
  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    Palette[0][i] = pListSort[i].uiData[0];
    Palette[1][i] = pListSort[i].uiData[1];
    Palette[2][i] = pListSort[i].uiData[2];
    if (pListSort[i].uiCnt == 0  )
    {
      uiPLTSize = i;
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;


}
#endif
Bool  TComPrediction::CalcPixelPredLuma(TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue, Pel * pEscapeFlag,   Pel*paRecoValue, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{
  Bool bLossless = pcCU->getCUTransquantBypass (0);
  const QpParam cQP(*pcCU, ComponentID(0));
  Int iQP = cQP.Qp;
  Int iQPrem = iQP%6;
  Int iQPper = iQP/6;
  Int quantiserScale =  g_quantScales[iQPrem];
  const Int quantiserRightShift = QUANT_SHIFT + iQPper;
  const Int rightShiftOffset = 1<< (quantiserRightShift-1);
  const Int InvquantiserRightShift = (IQUANT_SHIFT - iQPper);
  Int iAdd =  InvquantiserRightShift == 0? 0: 1 << (InvquantiserRightShift - 1);
  UInt uiY, uiX;
  uiY = uiIdx/uiWidth;
  uiX = uiIdx%uiWidth;

  UInt uiYIdxRaster = uiY*uiStrideOrg + uiX;
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);

  if (bLossless)
  {
    paRecoValue[uiYIdxRaster] = pValue[uiYIdxRaster] = pOrg[uiYIdxRaster];
  }
  else
  {
    pValue[uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[0]));
    paRecoValue[uiYIdxRaster] = (pValue[uiYIdxRaster] * g_invQuantScales[iQPrem] + iAdd) >> InvquantiserRightShift;
    paRecoValue[uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[uiYIdxRaster], g_bitDepth[0]));
  }

  return false; //do not use pixel prediction
}

Bool  TComPrediction::CalcPixelPredChroma(TComDataCU* pcCU, Pel* pOrg [2], Pel *pPalette[2], Pel* pValue[2], Pel * pEscapeFlag,   Pel*paRecoValue[2], UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{

  Bool bLossless = pcCU->getCUTransquantBypass (0);
  const QpParam cQP(*pcCU, ComponentID(1));//chroma
  Int iQP = cQP.Qp;
  Int iQPrem = iQP%6;
  Int iQPper = iQP/6;
  Int quantiserScale =  g_quantScales[iQPrem];
  const Int quantiserRightShift = QUANT_SHIFT + iQPper;
  const Int rightShiftOffset = 1<< (quantiserRightShift-1);
  const Int InvquantiserRightShift = (IQUANT_SHIFT - iQPper);
  Int iAdd =  InvquantiserRightShift == 0? 0: 1 << (InvquantiserRightShift - 1);
  UInt uiY, uiX;
  uiY = uiIdx/uiWidth;
  uiX = uiIdx%uiWidth;

  UInt uiYIdxRaster    = uiY*uiStrideOrg + uiX;
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);

  if (bLossless)
  {
    for (int i = 0; i < 2; i++)
    {
      pValue[i][uiYIdxRaster] = pOrg[i][uiYIdxRaster];
      paRecoValue[i][uiYIdxRaster] = pOrg[i][uiYIdxRaster];
    }
  }
  else
  {
    for (int i = 0; i < 2; i++)
    {
      pValue[i][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[i][uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[1]));
      paRecoValue[i][uiYIdxRaster] = (pValue[i][uiYIdxRaster] * g_invQuantScales[iQPrem] + iAdd) >> InvquantiserRightShift;
      paRecoValue[i][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[i][uiYIdxRaster], g_bitDepth[1]));
    }
  }

  return false; //do not use pixel prediction
}

Bool TComPrediction:: CalcPixelPred(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette[3], Pel* pValue, Pel * pEscapeFlag,  Pel*paPixelValue[3], Pel * paRecoValue[3], UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{
  Bool bLossless = pcCU->getCUTransquantBypass (0);
  Int iQP[3];
  Int iQPrem[3];
  Int iQPper[3];
  Int quantiserScale[3];
  Int quantiserRightShift[3];
  Int rightShiftOffset[3];
  Int InvquantiserRightShift[3];
  Int iAdd[3];
  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
  {
    QpParam cQP(*pcCU, ComponentID(ch));
    iQP[ch] = cQP.Qp;
    iQPrem[ch] = iQP[ch] % 6;
    iQPper[ch] = iQP[ch] / 6;
    quantiserScale[ch] = g_quantScales[iQPrem[ch]];
    quantiserRightShift[ch] = QUANT_SHIFT + iQPper[ch];
    rightShiftOffset[ch] = 1 << (quantiserRightShift[ch] - 1);
    InvquantiserRightShift[ch] = (IQUANT_SHIFT - iQPper[ch]);
    iAdd[ch] = InvquantiserRightShift[ch] == 0 ? 0 : 1 << (InvquantiserRightShift[ch] - 1);
  }

  UInt uiY, uiX;
  uiY = uiIdx / uiWidth;
  uiX = uiIdx % uiWidth;
  UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);

  if (bLossless)
  {
    for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
    {
      paPixelValue[ch][uiYIdxRaster] =  pOrg[ch][uiYIdxRaster];
      paRecoValue[ch][uiYIdxRaster] = pOrg[ch][uiYIdxRaster];
    }
  }
  else
  {
    for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
    {
      paPixelValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[ch][uiYIdxRaster] * quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch], uiMaxBit[ch]));
      paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch] [uiYIdxRaster]*g_invQuantScales[iQPrem[ch]] + iAdd[ch])>>InvquantiserRightShift[ch];
      paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch? 1:0]));
    }
  }

  return false; //do not use pixel prediction
}

#if PLT_IMPROVE_GEN
Bool  TComPrediction::CalCopyLuma(TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  //Bool bLossless = pcCU->getCUTransquantBypass(0);
  //UInt uiErrorThres = bLossless ? 0 : getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0;
  //, uiX = 0;
  //UInt uiBestIdx, uiMinError;
  Bool valid = false;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    //uiX = uiIdx%uiWidth;
    if (uiY == 0)
    {
      return false;
    }

    //UInt uiYIdxRaster    = uiY*uiStrideOrg + uiX;
    //UInt uiPLTIdx = 0;

    //uiMinError = MAX_UINT;
    //while (uiPLTIdx < uiPLTSize)
    //{
    //  UInt uiAbsError = abs (pPalette[uiPLTIdx] - pOrg[uiYIdxRaster]) ;

    //  if (uiAbsError < uiMinError)
    //  {
    //    uiBestIdx = uiPLTIdx;
    //    uiMinError = uiAbsError;
    //    pValue[uiIdx] = uiBestIdx;
    //    if (uiMinError == 0)
    //    {
    //      break;
    //    }
    //  }
    //  uiPLTIdx ++;
    //}

    UInt uiStride = uiWidth;
    pValue[uiIdx] = m_cIndexBlock[uiIdx] < 0 ? m_cIndexBlock[uiIdx] + MAX_PLT_SIZE : m_cIndexBlock[uiIdx];
    Bool bMismatch = (m_cIndexBlock[uiIdx] < 0);
    if (bMismatch && uiRun == 0)
    {
      return false;
    }

    if (uiRun == 0 && bMismatch)
    {
      return false;
    }
    if (pValue[uiIdx] == pValue[uiIdx - uiStride] && pEscapeFlag[uiIdx - uiStride] == 0  && bMismatch == 0)
    {
      uiRun ++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}

Bool TComPrediction::CalCopyChroma(TComDataCU *pcCU, Pel *pOrg [2], Pel *pPalette [2],  Pel *pValue, Pel *pEscapeFlag, Pel *pRun, UInt uiWidth, UInt uiHeight,
                                   UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  //Bool bLossless = pcCU->getCUTransquantBypass(0);
  //UInt uiErrorThres = bLossless ? 0 : 2 * getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0;
  //uiX = 0;
  //UInt uiBestIdx, uiMinError;
  Bool valid = false;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    //uiX = uiIdx%uiWidth;
    if (uiY == 0)
    {
      return false;
    }

    //UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    //UInt uiPLTIdx = 0;

    //uiMinError = MAX_UINT;
    //while (uiPLTIdx < uiPLTSize)
    //{
    //  UInt uiAbsError = abs (pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]) + abs (pPalette[1][uiPLTIdx] - pOrg[1][uiYIdxRaster]) ;
   
    //  if (uiAbsError < uiMinError)
    //  {
    //    uiBestIdx = uiPLTIdx;
    //    uiMinError = uiAbsError;
    //    pValue[uiIdx] = uiBestIdx;
    //    if (uiMinError == 0)
    //    {
    //      break;
    //    }
    //  }
    //  uiPLTIdx ++;
    //}

    //UInt uiStride = 0;

    UInt uiStride = uiWidth ;
    pValue[uiIdx] = m_cIndexBlockChroma[uiIdx] < 0 ? m_cIndexBlockChroma[uiIdx] + MAX_PLT_SIZE : m_cIndexBlockChroma[uiIdx];
    Bool bMismatch = (m_cIndexBlockChroma[uiIdx] < 0);
    if (bMismatch && uiRun == 0)
    {
      return false;
    }

    if (uiRun == 0 && bMismatch)
    {
      return false;
    }

    if (pValue[uiIdx] == pValue[uiIdx - uiStride] && pEscapeFlag[uiIdx - uiStride] == 0  && bMismatch == 0)
    {
      uiRun ++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}


Bool  TComPrediction::CalCopy(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette[3],  Pel* pValue, Pel *pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  UInt uiIdx = uiStartPos;
  UInt uiY = 0; 
  Bool valid = false;
  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    if (uiY == 0)
    {
      return false;
    }
    UInt uiStride = uiWidth;

    pValue[uiIdx] = m_cIndexBlock[uiIdx] < 0 ? m_cIndexBlock[uiIdx] + MAX_PLT_SIZE : m_cIndexBlock[uiIdx];
    Bool bMismatch = (m_cIndexBlock[uiIdx] < 0);

    if (uiRun == 0 && bMismatch)
    {
      return false;
    }
    if ((pValue[uiIdx] == pValue[uiIdx - uiStride]) && (pEscapeFlag[uiIdx - uiStride] == 0) && (bMismatch == 0))
    {
      uiRun ++;
      valid = true;
    }
    else
    {
      break;
    }
    uiIdx++;
  }
  return valid;
}

Bool TComPrediction::CalRunLuma (TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiErrorThres = bLossless ? 0 : getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY=0, uiX=0;
  UInt uiBestIdx, uiMinError;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;

    UInt uiPLTIdx = 0;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    { 
      UInt uiAbsError = abs (pPalette[uiPLTIdx] - pOrg[uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    Bool bMismatch = uiMinError > uiErrorThres;
#if PLT_FIX_CAL_RUN
    if (bMismatch && uiIdx == uiStartPos)
#else
    if (bMismatch && uiRun == 0)
#endif
    {
      return false;
    }

    if ( (uiIdx && (pSPoint[uiIdx-1]==0 && pEscapeFlag[uiIdx-1]==0)&&pValue[uiIdx]==pValue[uiIdx-1]) && bMismatch == 0)
    {
      uiRun ++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}


Bool TComPrediction::CalRunChroma(TComDataCU* pcCU, Pel* pOrg [2], Pel *pPalette [2], Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiErrorThres = bLossless ? 0 : 2 * getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0;
  UInt uiBestIdx, uiMinError;

  while (uiIdx < uiTotal)
  {

    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;

    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    { 

      UInt uiAbsError = abs(pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]) + abs(pPalette[1][uiPLTIdx] - pOrg[1][uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx ++;
    }

    Bool bMismatch = uiMinError > uiErrorThres;
#if PLT_FIX_CAL_RUN
    if (bMismatch && uiIdx == uiStartPos)
#else
    if (bMismatch && uiRun == 0)
#endif
    {
      return false;
    }

    if (uiIdx && (pSPoint[uiIdx - 1] == 0) && (pEscapeFlag[uiIdx - 1] == 0) 
              && (pValue[uiIdx] == pValue[uiIdx - 1]) && (bMismatch == 0))
    {
      uiRun ++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;

}

Bool TComPrediction::CalRun(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette [3] , Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{  

  UInt uiIdx = uiStartPos;

  while (uiIdx < uiTotal)
  {
    pValue[uiIdx] = m_cIndexBlock[uiIdx] < 0 ? m_cIndexBlock[uiIdx] + MAX_PLT_SIZE : m_cIndexBlock[uiIdx];
    Bool bMismatch = (m_cIndexBlock[uiIdx] < 0);
#if PLT_FIX_CAL_RUN
    if (bMismatch && uiIdx == uiStartPos)
#else
    if (bMismatch && uiRun == 0)
#endif
    {
      return false;
    }

    if ( (uiIdx && (pSPoint[uiIdx-1]==0 && pEscapeFlag[uiIdx-1]==0)&&pValue[uiIdx]==pValue[uiIdx-1]) && bMismatch == 0)
    {
      uiRun ++;
    }
    else if (uiIdx> uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}

#else
Bool  TComPrediction::CalCopyLuma(TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight, UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiErrorThres = bLossless ? 0 : getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0;
  UInt uiBestIdx, uiMinError;
  Bool valid = false;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx%uiWidth;
    if (uiY == 0)
    {
      return false;
    }

    UInt uiYIdxRaster = uiY*uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;

    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs(pPalette[uiPLTIdx] - pOrg[uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    UInt uiStride = uiWidth;


    Bool bMismatch = uiMinError>uiErrorThres;
    if (bMismatch && uiRun == 0)
      return false;

    if (uiRun == 0 && bMismatch)
      return false;

    if (pValue[uiIdx] == pValue[uiIdx - uiStride] && pEscapeFlag[uiIdx - uiStride] == 0 && bMismatch == 0)
    {
      uiRun++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}

Bool TComPrediction::CalCopyChroma(TComDataCU* pcCU, Pel* pOrg[2], Pel *pPalette[2], Pel* pValue, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight, UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiErrorThres = bLossless ? 0 : 2 * getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0;
  UInt uiBestIdx, uiMinError;
  Bool valid = false;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx%uiWidth;
    if (uiY == 0)
    {
      return false;
    }


    UInt uiYIdxRaster = uiY*uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;

    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs(pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]) + abs(pPalette[1][uiPLTIdx] - pOrg[1][uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    UInt uiStride = 0;

    uiStride = uiWidth;
    Bool bMismatch = uiMinError>uiErrorThres;
    if (bMismatch && uiRun == 0)
      return false;

    if (uiRun == 0 && bMismatch)
      return false;

    if (pValue[uiIdx] == pValue[uiIdx - uiStride] && pEscapeFlag[uiIdx - uiStride] == 0 && bMismatch == 0)
    {
      uiRun++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}

Bool  TComPrediction::CalCopy(TComDataCU* pcCU, Pel* pOrg[3], Pel *pPalette[3], Pel* pValue, Pel *pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight, UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);

  UInt uiYCbCrShift = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
  UInt uiXCbCrShift = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);

  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0, uiYCbCr = 0, uiXCbCr = 0;
  UInt uiStrideOrgCbCr = uiStrideOrg >> uiXCbCrShift;
  UInt uiBestIdx, uiMinError;
  ChromaFormat cCF = pcCU->getPic()->getSlice(0)->getSPS()->getChromaFormatIdc();
  Bool testC = true;
  Bool valid = false;
  Int iPLTErrLimit = bLossless ? 0 : getPLTErrLimit();
  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;
    if (uiY == 0)
    {
      return false;
    }

    testC = (((cCF == CHROMA_420) && (uiY % 2 || uiX % 2))
      || ((cCF == CHROMA_422) && (uiX % 2))) ? false : true;
    uiYCbCr = uiY >> uiYCbCrShift;
    uiXCbCr = uiX >> uiXCbCrShift;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiCbCrIdxRaster = uiYCbCr * uiStrideOrgCbCr + uiXCbCr;
    UInt uiPLTIdx = 0;

    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs(pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]);

      if (testC)
      {
        uiAbsError = uiAbsError + abs(pPalette[1][uiPLTIdx] - pOrg[1][uiCbCrIdxRaster]) + abs(pPalette[2][uiPLTIdx] - pOrg[2][uiCbCrIdxRaster]);
      }
      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    UInt uiStride = 0;

    uiStride = uiWidth;
    Bool bMismatch; //lossless only
    if (bLossless)
    {
      bMismatch = uiMinError>0;
    }
    else
    {
      if (testC)
      {
        bMismatch = uiMinError > 3 * iPLTErrLimit;
      }
      else
      {
        bMismatch = uiMinError > iPLTErrLimit;
      }
    }

    if (uiRun == 0 && bMismatch)
    {
      return false;
    }
    if ((pValue[uiIdx] == pValue[uiIdx - uiStride]) && (pEscapeFlag[uiIdx - uiStride] == 0) && (bMismatch == 0))
    {
      uiRun++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}

Bool TComPrediction::CalRunLuma(TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue, Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight, UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiErrorThres = bLossless ? 0 : getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0;
  UInt uiBestIdx, uiMinError;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;

    UInt uiPLTIdx = 0;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs(pPalette[uiPLTIdx] - pOrg[uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    Bool bMismatch = uiMinError > uiErrorThres;
#if PLT_FIX_CAL_RUN
    if (bMismatch && uiIdx == uiStartPos)
#else
    if (bMismatch && uiRun == 0)
#endif
    {
      return false;
    }

    if ((uiIdx && (pSPoint[uiIdx - 1] == 0 && pEscapeFlag[uiIdx - 1] == 0) && pValue[uiIdx] == pValue[uiIdx - 1]) && bMismatch == 0)
    {
      uiRun++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}


Bool TComPrediction::CalRunChroma(TComDataCU* pcCU, Pel* pOrg[2], Pel *pPalette[2], Pel* pValue, Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight, UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiErrorThres = bLossless ? 0 : 2 * getPLTErrLimit();
  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0;
  UInt uiBestIdx, uiMinError;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;

    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {

      UInt uiAbsError = abs(pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]) + abs(pPalette[1][uiPLTIdx] - pOrg[1][uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    Bool bMismatch = uiMinError > uiErrorThres;
#if PLT_FIX_CAL_RUN
    if (bMismatch && uiIdx == uiStartPos)
#else
    if (bMismatch && uiRun == 0)
#endif
    {
      return false;
    }

    if (uiIdx && (pSPoint[uiIdx - 1] == 0) && (pEscapeFlag[uiIdx - 1] == 0)
      && (pValue[uiIdx] == pValue[uiIdx - 1]) && (bMismatch == 0))
    {
      uiRun++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}

Bool TComPrediction::CalRun(TComDataCU* pcCU, Pel* pOrg[3], Pel *pPalette[3], Pel* pValue, Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight, UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiYCbCrShift = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
  UInt uiXCbCrShift = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);

  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0, uiYCbCr = 0, uiXCbCr = 0;
  UInt uiStrideOrgCbCr = uiStrideOrg >> uiXCbCrShift;
  UInt uiBestIdx, uiMinError;
  ChromaFormat cCF = pcCU->getPic()->getSlice(0)->getSPS()->getChromaFormatIdc();
  Bool testC = true;
  Int iPLTErrLimit = bLossless ? 0 : getPLTErrLimit();

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;

    uiYCbCr = uiY >> uiYCbCrShift;
    uiXCbCr = uiX >> uiXCbCrShift;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiCbCrIdxRaster = uiYCbCr * uiStrideOrgCbCr + uiXCbCr;
    UInt uiPLTIdx = 0;
    testC = (((cCF == CHROMA_420) && (uiY % 2 || uiX % 2)) ||
      ((cCF == CHROMA_422) && (uiX % 2))) ? false : true;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs(pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]);
      if (testC)
      {
        uiAbsError = uiAbsError + abs(pPalette[1][uiPLTIdx] - pOrg[1][uiCbCrIdxRaster]) + abs(pPalette[2][uiPLTIdx] - pOrg[2][uiCbCrIdxRaster]);
      }
      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    Bool bMismatch;
    if (bLossless)
    {
      bMismatch = uiMinError>0;
    }
    else
    {
      if (testC)
      {
        bMismatch = uiMinError> 3 * iPLTErrLimit;
      }
      else
      {
        bMismatch = uiMinError> iPLTErrLimit;
      }
    }
#if PLT_FIX_CAL_RUN
    if (bMismatch && uiIdx == uiStartPos)
#else
    if (bMismatch && uiRun == 0)
#endif
    {
      return false;
    }

    if ((uiIdx && (pSPoint[uiIdx - 1] == 0 && pEscapeFlag[uiIdx - 1] == 0) && pValue[uiIdx] == pValue[uiIdx - 1]) && bMismatch == 0)
    {
      uiRun++;
    }
    else if (uiIdx> uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}
#endif
#endif

//! \}
